@@ -4,7 +4,15 @@
 
 import os, time
 from cStringIO import StringIO
-from PyrexTypes import CPtrType
+from PyrexTypes import CPtrType, py_object_type, typecast
+from Pyrex.Utils import set
+
+#  Following is set by Testing.py to suppress filename/date comments
+#  in generated files, so as not to produce spurious changes in test
+#  reference files.
+
+testing_mode = False 
+                    
 
 import Code
 import Naming
@@ -34,11 +42,11 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         env.check_c_classes()
         self.body.analyse_expressions(env)
         env.return_type = PyrexTypes.c_void_type
-        self.referenced_modules = []
-        self.find_referenced_modules(env, self.referenced_modules, {})
+        self.referenced_modules = self.find_referenced_modules(env)
         if self.has_imported_c_functions():
             self.module_temp_cname = env.allocate_temp_pyobject()
             env.release_temp(self.module_temp_cname)
+        self.generate_dep_file(env, result)
         self.generate_c_code(env, result)
         self.generate_h_code(env, options, result)
         self.generate_api_code(env, result)
@@ -50,14 +58,34 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                     return 1
         return 0
     
+    def generate_dep_file(self, env, result):
+        modules = self.referenced_modules
+        includes = set(env.pyrex_include_files)
+        for module in modules:
+            for include in module.pyrex_include_files:
+                includes.add(include)
+        if len(modules) > 1 or includes:
+            include_list = list(includes)
+            include_list.sort()
+            dep_file = replace_suffix(result.c_file, ".dep")
+            f = open(dep_file, "w")
+            try:
+                for module in modules[:-1]:
+                    f.write("cimport %s\n" % module.qualified_name)
+                for path in include_list:
+                    f.write("include %s\n" % path)
+            finally:
+                f.close()
+    
     def generate_h_code(self, env, options, result):
-        def h_entries(entries, pxd = 0):
+        def pub(entries): #, pxd = 0):
             return [entry for entry in entries
-                if entry.visibility == 'public' or pxd and entry.defined_in_pxd]
-        h_types = h_entries(env.type_entries)
-        h_vars = h_entries(env.var_entries)
-        h_funcs = h_entries(env.cfunc_entries)
-        h_extension_types = h_entries(env.c_class_entries)
+                if entry.visibility == 'public'] # or pxd and entry.defined_in_pxd]
+        denv = env.definition_scope
+        h_types = pub(denv.type_entries) + pub(env.type_entries)
+        h_vars = pub(denv.var_entries) + pub(env.var_entries)
+        h_funcs = pub(denv.cfunc_entries) + pub(env.cfunc_entries)
+        h_extension_types = pub(denv.c_class_entries) + pub(env.c_class_entries)
         if h_types or h_vars or h_funcs or h_extension_types:
             result.h_file = replace_suffix(result.c_file, ".h")
             h_code = Code.CCodeWriter(open_new_file(result.h_file))
@@ -106,13 +134,14 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         return env.qualified_name.replace(".", "__")
     
     def generate_api_code(self, env, result):
+        denv = env.definition_scope
         api_funcs = []
         public_extension_types = []
         has_api_extension_types = 0
-        for entry in env.cfunc_entries:
+        for entry in denv.cfunc_entries + env.cfunc_entries:
             if entry.api:
                 api_funcs.append(entry)
-        for entry in env.c_class_entries:
+        for entry in denv.c_class_entries + env.c_class_entries:
             if entry.visibility == 'public':
                 public_extension_types.append(entry)
             if entry.api:
@@ -160,7 +189,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                         entry.name,
                         entry.cname,
                         sig))
-            h_code.putln("Py_DECREF(module);")
+            h_code.putln("Py_DECREF(module); module = 0;")
             for entry in public_extension_types:
                 self.generate_type_import_call(entry.type, h_code, "goto bad;")
             h_code.putln("return 0;")
@@ -191,30 +220,43 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         i_code.dedent()
     
     def generate_c_code(self, env, result):
-        modules = self.referenced_modules
-        code = Code.CCodeWriter(StringIO())
+        code = Code.MainCCodeWriter(StringIO())
         code.h = Code.CCodeWriter(StringIO())
         code.init_labels()
+        
+        modules = self.referenced_modules
         self.generate_module_preamble(env, modules, code.h)
 
         code.putln("")
         code.putln("/* Implementation of %s */" % env.qualified_name)
-        self.generate_const_definitions(env, code)
-        self.generate_interned_name_decls(env, code)
-        self.generate_py_string_decls(env, code)
+        #self.generate_const_definitions(env, code)
+        #self.generate_interned_name_decls(env, code)
+        #self.generate_py_string_decls(env, code)
         self.body.generate_function_definitions(env, code)
-        self.generate_interned_name_table(env, code)
-        self.generate_py_string_table(env, code)
+        #self.generate_interned_name_table(env, code)
+        #self.generate_py_string_table(env, code)
         self.generate_typeobj_definitions(env, code)
         self.generate_method_table(env, code)
         self.generate_filename_init_prototype(code)
         self.generate_module_init_func(modules[:-1], env, code)
         self.generate_filename_table(code)
-        self.generate_utility_functions(env, code)
+        self.generate_utility_functions(code)
 
+        denv = env.definition_scope
         for module in modules:
+            code.h.putln("")
+            code.h.putln("/* Declarations from %s */" % module.qualified_name)
             self.generate_declarations_for_module(module, code.h,
-                definition = module is env)
+                implementation = module is denv)
+
+        code.h.putln("")
+        code.h.putln("/* Declarations from implementation of %s */" %
+            env.qualified_name)
+        self.generate_declarations_for_module(env, code.h, implementation = 1)
+        code.global_state.generate_const_declarations(code.h)
+        #self.generate_interned_name_table(code.interned_strings, code.h)
+        #self.generate_py_string_table(code.py_strings, code.h)
+        self.generate_default_value_declarations(env, code.h)
 
         f = open_new_file(result.c_file)
         f.write(code.h.f.getvalue())
@@ -223,16 +265,39 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         f.close()
         result.c_file_generated = 1
     
-    def find_referenced_modules(self, env, module_list, modules_seen):
-        if env not in modules_seen:
-            modules_seen[env] = 1
-            for imported_module in env.cimported_modules:
-                self.find_referenced_modules(imported_module, module_list, modules_seen)
-            module_list.append(env)
+    def find_referenced_modules(self, env):
+        #  Given the ImplementationScope, find the DefinitionScopes of all
+        #  modules cimported, directly or indirectly. Includes this module's
+        #  DefinitionScope as the last entry in the list.
+        denv = env.definition_scope
+        module_list = []
+        modules_seen = set()
+        def add_module(module):
+            if module not in modules_seen:
+                modules_seen.add(module)
+                add_modules(module.cimported_modules)
+                module_list.append(module)
+        def add_modules(modules):
+            for module in modules:
+                add_module(module)
+        modules_seen.add(denv)
+        add_modules(denv.cimported_modules)
+        add_modules(env.cimported_modules)
+        module_list.append(denv)
+        #self.print_referenced_modules(module_list) ###
+        return module_list
+    
+    def print_referenced_modules(self, module_list):
+        print "find_referenced_modules: result =",
+        for m in module_list:
+            print m,
+        print
         
     def generate_module_preamble(self, env, cimported_modules, code):
-        code.putln('/* Generated by Pyrex %s on %s */' % (
-            Version.version, time.asctime()))
+        comment = "Generated by Pyrex"
+        if not testing_mode:
+            comment = "%s %s on %s" % (comment, Version.version, time.asctime())
+        code.putln('/* %s */' % comment)
         code.putln('')
         code.putln('#define PY_SSIZE_T_CLEAN')
         for filename in env.python_include_files:
@@ -248,8 +313,12 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln("  #define PyInt_AsSsize_t(o)	PyInt_AsLong(o)")
         code.putln("#endif")
         code.putln("#ifndef WIN32")
-        code.putln("  #define __stdcall")
-        code.putln("  #define __cdecl")
+        code.putln("  #ifndef __stdcall")
+        code.putln("    #define __stdcall")
+        code.putln("  #endif")
+        code.putln("  #ifndef __cdecl")
+        code.putln("    #define __cdecl")
+        code.putln("  #endif")
         code.putln("#endif")
         self.generate_extern_c_macro_definition(code)
         code.putln("#include <math.h>")
@@ -262,9 +331,16 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln('static int %s;' % Naming.lineno_cname)
         code.putln('static char *%s;' % Naming.filename_cname)
         code.putln('static char **%s;' % Naming.filetable_cname)
-        if env.doc:
+        doc = None
+        doc1 = env.definition_scope.doc
+        doc2 = env.doc
+        if doc1 and doc2:
+            doc = "%s\\n%s" % (doc1, doc2)
+        else:
+            doc = doc1 or doc2
+        if doc:
             code.putln('')
-            code.putln('static char %s[] = "%s";' % (env.doc_cname, env.doc))
+            code.putln('static char %s[] = "%s";' % (env.doc_cname, doc))
     
     def generate_extern_c_macro_definition(self, code):
         name = Naming.extern_c_macro
@@ -284,26 +360,13 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             code.putln('#include "%s"' % filename)
     
     def generate_filename_table(self, code):
-        code.putln("")
-        code.putln("static char *%s[] = {" % Naming.filenames_cname)
-        if code.filename_list:
-            for filename in code.filename_list:
-                filename = os.path.basename(filename)
-                escaped_filename = filename.replace("\\", "\\\\").replace('"', r'\"')
-                code.putln('"%s",' % 
-                    escaped_filename)
-        else:
-            # Some C compilers don't like an empty array
-            code.putln("0")
-        code.putln("};")
+        code.global_state.generate_filename_table(code)
     
-    def generate_declarations_for_module(self, env, code, definition):
-        code.putln("")
-        code.putln("/* Declarations from %s */" % env.qualified_name)
+    def generate_declarations_for_module(self, env, code, implementation):
         self.generate_type_predeclarations(env, code)
-        self.generate_type_definitions(env, code, definition)
-        self.generate_global_declarations(env, code, definition)
-        self.generate_cfunction_predeclarations(env, code, definition)
+        self.generate_type_definitions(env, code) #, implementation)
+        self.generate_global_declarations(env, code, implementation)
+        self.generate_cfunction_predeclarations(env, code, implementation)
 
     def generate_type_predeclarations(self, env, code):
         pass
@@ -325,14 +388,16 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 elif type.is_extension_type:
                     self.generate_obj_struct_definition(type, code)
     
-    def generate_type_definitions(self, env, code, definition):
-        if definition:
-            type_entries = env.type_entries
-        else:
-            type_entries = []
-            for entry in env.type_entries:
-                if entry.defined_in_pxd:
-                    type_entries.append(entry)
+    def generate_type_definitions(self, env, code): #, implementation):
+        #print "generate_type_definitions:", env ###
+#		if definition:
+#			type_entries = env.type_entries
+#		else:
+#			type_entries = []
+#			for entry in env.type_entries:
+#				if entry.defined_in_pxd:
+#					type_entries.append(entry)
+        type_entries = env.type_entries
         self.generate_type_header_code(type_entries, code)
         for entry in env.c_class_entries:
             if not entry.in_cinclude:
@@ -342,17 +407,18 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
     
     def generate_gcc33_hack(self, env, code):
         # Workaround for spurious warning generation in gcc 3.3
-        code.putln("")
-        for entry in env.c_class_entries:
-            type = entry.type
-            if not type.typedef_flag:
-                name = type.objstruct_cname
-                if name.startswith("__pyx_"):
-                    tail = name[6:]
-                else:
-                    tail = name
-                code.putln("typedef struct %s __pyx_gcc33_%s;" % (
-                    name, tail))
+        if 0:
+            code.putln("")
+            for entry in env.c_class_entries:
+                type = entry.type
+                if not type.typedef_flag:
+                    name = type.objstruct_cname
+                    if name.startswith("__pyx_"):
+                        tail = name[6:]
+                    else:
+                        tail = name
+                    code.putln("typedef struct %s __pyx_gcc33_%s;" % (
+                        name, tail))
     
     def generate_typedef(self, entry, code):
         base_type = entry.type.typedef_base_type
@@ -377,7 +443,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             code.putln("")
             code.putln(header)
             var_entries = scope.var_entries
-            if not var_entries:
+            if not var_entries and not scope.cfunc_entries:
                 error(entry.pos,
                     "Empty struct or union definition not allowed outside a"
                     " 'cdef extern from' block")
@@ -488,26 +554,32 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                     attr.type.declaration_code(attr.cname))
         code.putln(footer)
     
-    def generate_global_declarations(self, env, code, definition):
+    def generate_global_declarations(self, env, code, implementation):
         code.putln("")
         for entry in env.c_class_entries:
-            if definition or entry.defined_in_pxd:
+            if implementation or entry.defined_in_pxd:
                 code.putln("static PyTypeObject *%s = 0;" % 
                     entry.type.typeptr_cname)
+        #code.putln("/* var_entries */") ###
         code.put_var_declarations(env.var_entries, static = 1, 
-            dll_linkage = "DL_EXPORT", definition = definition)
+            dll_linkage = "DL_EXPORT", definition = implementation)
+    
+    def generate_default_value_declarations(self, env, code):
+        #code.putln("/* default_entries */") ###
+        code.putln("")
         code.put_var_declarations(env.default_entries, static = 1)
     
-    def generate_cfunction_predeclarations(self, env, code, definition):
+    def generate_cfunction_predeclarations(self, env, code, implementation):
         for entry in env.cfunc_entries:
-            if not entry.in_cinclude and (definition
-                    or entry.defined_in_pxd or entry.visibility == 'extern'):
+            if not entry.in_cinclude:
+                    # and (definition or entry.defined_in_pxd or
+                    # entry.visibility == 'extern'):
                 if entry.visibility in ('public', 'extern'):
                     dll_linkage = "DL_EXPORT"
                 else:
                     dll_linkage = None
                 type = entry.type
-                if not definition and entry.defined_in_pxd:
+                if not implementation: #and entry.defined_in_pxd:
                     type = CPtrType(type)
                 header = type.declaration_code(entry.cname, 
                     dll_linkage = dll_linkage)
@@ -521,7 +593,8 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
     
     def generate_typeobj_definitions(self, env, code):
         full_module_name = env.qualified_name
-        for entry in env.c_class_entries:
+        denv = env.definition_scope
+        for entry in denv.c_class_entries + env.c_class_entries:
             #print "generate_typeobj_definitions:", entry.name
             #print "...visibility =", entry.visibility
             if entry.visibility <> 'extern':
@@ -628,10 +701,11 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln(
             "static void %s(PyObject *o) {"
                 % scope.mangle_internal("tp_dealloc"))
-        py_attrs = []
-        for entry in scope.var_entries:
-            if entry.type.is_pyobject and entry.name <> "__weakref__":
-                py_attrs.append(entry)
+        #py_attrs = []
+        #for entry in scope.var_entries:
+        #	if entry.type.is_pyobject and entry.name <> "__weakref__":
+        #		py_attrs.append(entry)
+        py_attrs = scope.pyattr_entries
         if py_attrs:
             self.generate_self_cast(scope, code)
         self.generate_usr_dealloc_call(scope, code)
@@ -673,65 +747,77 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 "}")
     
     def generate_traverse_function(self, scope, code):
-        base_type = scope.parent_type.base_type
-        code.putln("")
-        code.putln(
-            "static int %s(PyObject *o, visitproc v, void *a) {"
-                % scope.mangle_internal("tp_traverse"))
-        py_attrs = []
-        for entry in scope.var_entries:
-            if entry.type.is_pyobject and entry.name <> "__weakref__":
-                py_attrs.append(entry)
-        if base_type or py_attrs:
-            code.putln(
-                    "int e;")
+        py_attrs = scope.pyattr_entries
         if py_attrs:
+            base_type = scope.parent_type.base_type
+            code.putln("")
+            code.putln(
+                "static int %s(PyObject *o, visitproc v, void *a) {"
+                    % scope.mangle_internal("tp_traverse"))
+            code.putln(
+                "int e;")
             self.generate_self_cast(scope, code)
-        if base_type:
-            code.putln(
-                    "e = %s->tp_traverse(o, v, a); if (e) return e;" %
-                        base_type.typeptr_cname)
-        for entry in py_attrs:
-            var_code = "p->%s" % entry.cname
-            code.putln(
-                    "if (%s) {"
-                        % var_code)
-            if entry.type.is_extension_type:
-                var_code = "((PyObject*)%s)" % var_code
-            code.putln(
-                        "e = (*v)(%s, a); if (e) return e;" 
+            if base_type:
+                code.putln(
+                    "traverseproc t;")
+                code.putln(
+                        "if ((t = %s->tp_traverse)) {" %
+                            base_type.typeptr_cname)
+                code.putln(
+                            "e = t(o, v, a); if (e) return e;")
+                code.putln(
+                        "}")
+            for entry in py_attrs:
+                var_code = "p->%s" % entry.cname
+                code.putln(
+                        "if (%s) {"
                             % var_code)
+                if entry.type.is_extension_type:
+                    var_code = "((PyObject*)%s)" % var_code
+                code.putln(
+                            "e = (*v)(%s, a); if (e) return e;" 
+                                % var_code)
+                code.putln(
+                        "}")
             code.putln(
-                    "}")
-        code.putln(
-                "return 0;")
-        code.putln(
-            "}")
+                    "return 0;")
+            code.putln(
+                "}")
     
     def generate_clear_function(self, scope, code):
-        base_type = scope.parent_type.base_type
-        code.putln("")
-        code.putln(
-            "static int %s(PyObject *o) {"
-                % scope.mangle_internal("tp_clear"))
-        py_attrs = []
-        for entry in scope.var_entries:
-            if entry.type.is_pyobject and entry.name <> "__weakref__":
-                py_attrs.append(entry)
+        py_attrs = scope.pyattr_entries
         if py_attrs:
-            self.generate_self_cast(scope, code)
-        if base_type:
+            base_type = scope.parent_type.base_type
+            code.putln("")
             code.putln(
-                "%s->tp_clear(o);" %
-                    base_type.typeptr_cname)
-        for entry in py_attrs:
-            name = "p->%s" % entry.cname
-            code.put_xdecref(name, entry.type)
-            code.put_init_var_to_py_none(entry, "p->%s")
-        code.putln(
-            "return 0;")
-        code.putln(
-            "}")
+                "static int %s(PyObject *o) {"
+                    % scope.mangle_internal("tp_clear"))
+            self.generate_self_cast(scope, code)
+            code.putln(
+                    "PyObject *t;")
+            if base_type:
+                code.putln(
+                    "inquiry c;")
+                code.putln(
+                    "if ((c = %s->tp_clear)) {" %
+                        base_type.typeptr_cname)
+                code.putln(
+                        "c(o);")
+                code.putln(
+                    "}")
+            for entry in py_attrs:
+                name = "p->%s" % entry.cname
+                code.putln(
+                    "t = %s; " %
+                        typecast(py_object_type, entry.type, name))
+                code.put_init_var_to_py_none(entry, "p->%s")
+                #code.put_xdecref(name, entry.type)
+                code.putln(
+                    "Py_XDECREF(t);")
+            code.putln(
+                "return 0;")
+            code.putln(
+                "}")
         
     def generate_getitem_int_function(self, scope, code):
         # This function is put into the sq_item slot when
@@ -1076,9 +1162,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         if entry.visibility == 'public':
             header = "DL_EXPORT(PyTypeObject) %s = {"
         else:
-            #header = "statichere PyTypeObject %s = {"
             header = "PyTypeObject %s = {"
-        #code.putln(header % scope.parent_type.typeobj_cname)
         code.putln(header % type.typeobj_cname)
         code.putln(
             "PyObject_HEAD_INIT(0)")
@@ -1149,58 +1233,97 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 "static struct PyGetSetDef %s[] = {" %
                     env.getset_table_cname)
             for entry in env.property_entries:
+                if entry.doc:
+                    doc_code = code.get_string_const(entry.doc)
+                else:
+                    doc_code = "0"
                 code.putln(
                     '{"%s", %s, %s, %s, 0},' % (
                         entry.name,
                         entry.getter_cname or "0",
                         entry.setter_cname or "0",
-                        entry.doc_cname or "0"))
+                        doc_code))
             code.putln(
                     "{0, 0, 0, 0, 0}")
             code.putln(
                 "};")
     
-    def generate_interned_name_table(self, env, code):
-        items = env.intern_map.items()
-        if items:
-            items.sort()
-            code.putln("")
-            code.putln(
-                "static __Pyx_InternTabEntry %s[] = {" %
-                    Naming.intern_tab_cname)
-            for (name, cname) in items:
-                code.putln(
-                    '{&%s, "%s"},' % (
-                        cname,
-                        name))
-            code.putln(
-                "{0, 0}")
-            code.putln(
-                "};")
+#	def generate_interned_name_table(self, env, code):
+#		items = env.intern_map.items()
+#		if items:
+#			items.sort()
+#			code.putln("")
+#			code.putln(
+#				"static __Pyx_InternTabEntry %s[] = {" %
+#					Naming.intern_tab_cname)
+#			for (name, cname) in items:
+#				code.putln(
+#					'{&%s, "%s"},' % (
+#						cname,
+#						name))
+#			code.putln(
+#				"{0, 0}")
+#			code.putln(
+#				"};")
     
-    def generate_py_string_table(self, env, code):
-        entries = env.all_pystring_entries
-        if entries:
-            code.putln("")
-            code.putln(
-                "static __Pyx_StringTabEntry %s[] = {" %
-                    Naming.stringtab_cname)
-            for entry in entries:
-                code.putln(
-                    "{&%s, %s, sizeof(%s)}," % (
-                        entry.pystring_cname,
-                        entry.cname,
-                        entry.cname))
-            code.putln(
-                "{0, 0, 0}")
-            code.putln(
-                "};")
+#	def generate_py_string_table(self, env, code):
+#		entries = env.all_pystring_entries
+#		if entries:
+#			code.putln("")
+#			code.putln(
+#				"static __Pyx_StringTabEntry %s[] = {" %
+#					Naming.stringtab_cname)
+#			for entry in entries:
+#				code.putln(
+#					"{&%s, %s, sizeof(%s)}," % (
+#						entry.pystring_cname,
+#						entry.cname,
+#						entry.cname))
+#			code.putln(
+#				"{0, 0, 0}")
+#			code.putln(
+#				"};")
     
+    def generate_interned_name_table(self, interned_strings, code):
+        code.putln("")
+        code.putln(
+            "static PyObject **%s[] = {" % Naming.intern_tab_cname)
+        for s in interned_strings:
+            code.putln("&%s," % s.py_cname)
+        code.putln("0")
+        code.putln(
+            "};")
+    
+#	def generate_py_string_table(self, py_strings, code):
+#		interned = []
+#		uninterned = []
+#		for s in py_strings:
+#			if s.is_interned:
+#				interned.append(s)
+#			else:
+#				uninterned.append(s)
+#		code.putln("")
+#		code.putln(
+#			"static __Pyx_StringTabEntry %s[] = {" %
+#				Naming.stringtab_cname)
+#		for s in interned + uninterned:
+#			cname = s.cname
+#			code.putln(
+#				"{&%s, %s, sizeof(%s)}," % (
+#					s.py_cname,
+#					cname,
+#					cname))
+#		code.putln(
+#			"{0, 0, 0}")
+#		code.putln(
+#			"};")
+
     def generate_filename_init_prototype(self, code):
         code.putln("");
         code.putln("static void %s(void); /*proto*/" % Naming.fileinit_cname)
 
     def generate_module_init_func(self, imported_modules, env, code):
+        denv = env.definition_scope
         code.putln("")
         header = "PyMODINIT_FUNC init%s(void)" % env.module_name
         code.putln("%s; /*proto*/" % header)
@@ -1214,17 +1337,18 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         #code.putln("/*--- Module creation code ---*/")
         self.generate_module_creation_code(env, code)
 
-        #code.putln("/*--- Intern code ---*/")
-        self.generate_intern_code(env, code)
-
         #code.putln("/*--- String init code ---*/")
         self.generate_string_init_code(env, code)
 
+        #code.putln("/*--- Intern code ---*/")
+        #self.generate_intern_code(env, code)
+
         #code.putln("/*--- Global init code ---*/")
         self.generate_global_init_code(env, code)
         
         #code.putln("/*--- Function export code ---*/")
-        self.generate_c_function_export_code(env, code)
+        self.generate_pxd_function_export_code(env, code)
+        self.generate_api_function_export_code(env, code)
 
         #code.putln("/*--- Function import code ---*/")
         for module in imported_modules:
@@ -1243,7 +1367,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.put_label(code.error_label)
         code.put_var_xdecrefs(env.temp_entries)
         code.putln('__Pyx_AddTraceback("%s");' % (env.qualified_name))
-        env.use_utility_code(Nodes.traceback_utility_code)
+        code.use_utility_code(Nodes.traceback_utility_code)
         code.putln('}')
     
     def generate_filename_init_call(self, code):
@@ -1266,6 +1390,9 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             "if (!%s) %s;" % (
                 env.module_cname,
                 code.error_goto(self.pos)));
+        code.putln(
+            "Py_INCREF(%s);" %
+                env.module_cname)
         code.putln(
             '%s = PyImport_AddModule("__builtin__");' %
                 Naming.builtins_cname)
@@ -1279,21 +1406,19 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 Naming.builtins_cname,
                 code.error_goto(self.pos)))
     
-    def generate_intern_code(self, env, code):
-        if env.intern_map:
-            env.use_utility_code(Nodes.init_intern_tab_utility_code);
-            code.putln(
-                "if (__Pyx_InternStrings(%s) < 0) %s;" % (
-                    Naming.intern_tab_cname,
-                    code.error_goto(self.pos)))
+#	def generate_intern_code(self, env, code):
+#		code.use_utility_code(Nodes.init_intern_tab_utility_code);
+#		code.putln(
+#			"if (__Pyx_InternStrings(%s) < 0) %s;" % (
+#				Naming.intern_tab_cname,
+#				code.error_goto(self.pos)))
     
     def generate_string_init_code(self, env, code):
-        if env.all_pystring_entries:
-            env.use_utility_code(Nodes.init_string_tab_utility_code)
-            code.putln(
-                "if (__Pyx_InitStrings(%s) < 0) %s;" % (
-                    Naming.stringtab_cname,
-                    code.error_goto(self.pos)))
+        code.use_utility_code(Nodes.init_string_tab_utility_code)
+        code.putln(
+            "if (__Pyx_InitStrings(%s) < 0) %s;" % (
+                Naming.stringtab_cname,
+                code.error_goto(self.pos)))
     
     def generate_global_init_code(self, env, code):
         # Generate code to initialise global PyObject *
@@ -1302,23 +1427,36 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             if entry.visibility <> 'extern':
                 if entry.type.is_pyobject:
                     code.put_init_var_to_py_none(entry)
-
-    def generate_c_function_export_code(self, env, code):
-        # Generate code to create PyCFunction wrappers for exported C functions.
+    
+    def generate_pxd_function_export_code(self, env, code):
+        denv = env.definition_scope
+        for entry in denv.cfunc_entries:
+            self.generate_c_function_export_code(env, entry, code)
+    
+    def generate_api_function_export_code(self, env, code):
         for entry in env.cfunc_entries:
-            if entry.api or entry.defined_in_pxd:
-                env.use_utility_code(function_export_utility_code)
-                signature = entry.type.signature_string()
-                code.putln('if (__Pyx_ExportFunction("%s", (void*)%s, "%s") < 0) %s' % (
-                    entry.name,
-                    entry.cname,
-                    signature, 
-                    code.error_goto(self.pos)))
+            if entry.api:
+                self.generate_c_function_export_code(env, entry, code)
+
+#	def generate_c_function_export_code(self, env, code):
+#		# Generate code to create PyCFunction wrappers for exported C functions.
+#		for entry in env.cfunc_entries:
+#			if entry.api or entry.defined_in_pxd:
+    
+    def generate_c_function_export_code(self, env, entry, code):
+        code.use_utility_code(function_export_utility_code)
+        signature = entry.type.signature_string()
+        code.putln('if (__Pyx_ExportFunction("%s", (void*)%s, "%s") < 0) %s' % (
+            entry.name,
+            entry.cname,
+            signature, 
+            code.error_goto(self.pos)))
     
     def generate_type_import_code_for_module(self, module, env, code):
         # Generate type import code for all exported extension types in
         # an imported module.
         #if module.c_class_entries:
+        #print "generate_type_import_code_for_module:", module ###
         for entry in module.c_class_entries:
             if entry.defined_in_pxd:
                 self.generate_type_import_code(env, entry.type, entry.pos, code)
@@ -1330,8 +1468,8 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             if entry.defined_in_pxd:
                 entries.append(entry)
         if entries:
-            env.use_utility_code(import_module_utility_code)
-            env.use_utility_code(function_import_utility_code)
+            code.use_utility_code(import_module_utility_code)
+            code.use_utility_code(function_import_utility_code)
             temp = self.module_temp_cname
             code.putln(
                 '%s = __Pyx_ImportModule("%s"); if (!%s) %s' % (
@@ -1352,7 +1490,9 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
     def generate_type_init_code(self, env, code):
         # Generate type import code for extern extension types
         # and type ready code for non-extern ones.
-        for entry in env.c_class_entries:
+        #print "generate_type_init_code:", env ###
+        denv = env.definition_scope
+        for entry in denv.c_class_entries + env.c_class_entries:
             if entry.visibility == 'extern':
                 self.generate_type_import_code(env, entry.type, entry.pos, code)
             else:
@@ -1366,15 +1506,16 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         if base_type and base_type.module_name <> env.qualified_name:
             self.generate_type_import_code(env, base_type, self.pos, code)
     
-    def use_type_import_utility_code(self, env):
+    def use_type_import_utility_code(self, code):
         import ExprNodes
-        env.use_utility_code(type_import_utility_code)
-        env.use_utility_code(import_module_utility_code)
+        code.use_utility_code(type_import_utility_code)
+        code.use_utility_code(import_module_utility_code)
     
     def generate_type_import_code(self, env, type, pos, code):
         # If not already done, generate code to import the typeobject of an
         # extension type defined in another module, and extract its C method
         # table pointer if any.
+        #print "generate_type_import_code:", type ###
         if type in env.types_imported:
             return
         if type.typedef_flag:
@@ -1382,14 +1523,14 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         else:
             objstruct = "struct %s" % type.objstruct_cname
         self.generate_type_import_call(type, code, code.error_goto(pos))
-        self.use_type_import_utility_code(env)
+        self.use_type_import_utility_code(code)
         if type.vtabptr_cname:
             code.putln(
                 "if (__Pyx_GetVtable(%s->tp_dict, &%s) < 0) %s" % (
                     type.typeptr_cname,
                     type.vtabptr_cname,
                     code.error_goto(pos)))
-            env.use_utility_code(Nodes.get_vtable_utility_code)
+            code.use_utility_code(Nodes.get_vtable_utility_code)
         env.types_imported[type] = 1
     
     def generate_type_import_call(self, type, code, error_code):
@@ -1425,7 +1566,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                             typeobj_cname,
                             type.vtabptr_cname,
                             code.error_goto(entry.pos)))
-                    env.use_utility_code(Nodes.set_vtable_utility_code)
+                    code.use_utility_code(Nodes.set_vtable_utility_code)
                 code.putln(
                     'if (PyObject_SetAttrString(%s, "%s", (PyObject *)&%s) < 0) %s' % (
                         Naming.module_cname,
@@ -1476,17 +1617,8 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 "%s = &%s;" % (
                     type.typeptr_cname, type.typeobj_cname))
     
-    def generate_utility_functions(self, env, code):
-        code.putln("")
-        code.putln("/* Runtime support code */")
-        code.putln("")
-        code.putln("static void %s(void) {" % Naming.fileinit_cname)
-        code.putln("%s = %s;" % 
-            (Naming.filetable_cname, Naming.filenames_cname))
-        code.putln("}")
-        for utility_code in env.utility_code_used:
-            code.h.put(utility_code[0])
-            code.put(utility_code[1])
+    def generate_utility_functions(self, code):
+        code.global_state.generate_utility_functions(code)
 
 #------------------------------------------------------------------------------------
 #
@@ -1498,6 +1630,8 @@ import_module_utility_code = [
 """
 static PyObject *__Pyx_ImportModule(char *name); /*proto*/
 ""","""
+#ifndef __PYX_HAVE_RT_ImportModule
+#define __PYX_HAVE_RT_ImportModule
 static PyObject *__Pyx_ImportModule(char *name) {
     PyObject *py_name = 0;
     
@@ -1509,6 +1643,7 @@ bad:
     Py_XDECREF(py_name);
     return 0;
 }
+#endif
 """]
 
 #------------------------------------------------------------------------------------
@@ -1617,11 +1752,9 @@ static int __Pyx_ImportFunction(PyObject *module, char *funcname, void **f, char
         goto bad;
     }
     *f = PyCObject_AsVoidPtr(cobj);
-    Py_DECREF(cobj);
     Py_DECREF(d);
     return 0;
 bad:
-    Py_XDECREF(cobj);
     Py_XDECREF(d);
     return -1;
 }
